#################################################################
# Importing samtools and htslib
#
# For htslib, simply copy the whole release tar-ball
# into the directory "htslib" and recreate the file version.h
#
# rm -rf htslib
# mv download/htslib htslib
# git checkout -- htslib/version.h
# Edit the file htslib/version.h to set the right version number.
#
# For samtools, type:
# rm -rf samtools
# python import.py samtools download/samtools
# git checkout -- samtools/version.h
#
# Manually, then:
# modify config.h to set compatibility flags
#
# For bcftools, type:
# rm -rf bcftools
# python import.py bcftools download/bedtools
# git checkout -- bcftools/version.h
# rm -rf bedtools/test bedtools/plugins

import fnmatch
import os
import re
import itertools
import shutil
import sys
import hashlib


EXCLUDE = {
    "samtools": (
        "razip.c",
        "bgzip.c",
        "main.c",
        "calDepth.c",
        "bam2bed.c",
        "wgsim.c",
        "bam_tview.c",
        "bam_tview.h",
        "bam_tview_html.c",
        "bam_tview_curses.c",
        "md5fa.c",
        "md5sum-lite.c",
        "maq2sam.c",
        "bamcheck.c",
        "chk_indel.c",
        "vcf-miniview.c",
        "hfile_irods.c",  # requires irods library
    ),
    "bcftools": (
        "test", "plugins", "peakfit.c",
        "peakfit.h",
        # needs to renamed, name conflict with samtools reheader
        # "reheader.c",
        "polysomy.c"),
    "htslib": (
        'htslib/tabix.c', 'htslib/bgzip.c',
        'htslib/htsfile.c', 'htslib/hfile_irods.c'),
}


MAIN = {
    "samtools": "bamtk",
    "bcftools": "main"
}



def locate(pattern, root=os.curdir):
    '''Locate all files matching supplied filename pattern in and below
    supplied root directory.
    '''
    for path, dirs, files in os.walk(os.path.abspath(root)):
        for filename in fnmatch.filter(files, pattern):
            yield os.path.join(path, filename)


def _update_pysam_files(cf, destdir):
    '''update pysam files applying redirection of ouput'''
    basename = os.path.basename(destdir)
    for filename in cf:
        if not filename:
            continue
        dest = filename + ".pysam.c"
        with open(filename, encoding="utf-8") as infile:
            lines = "".join(infile.readlines())

            with open(dest, "w", encoding="utf-8") as outfile:
                outfile.write('#include "{}.pysam.h"\n\n'.format(basename))
                subname, _ = os.path.splitext(os.path.basename(filename))
                if subname in MAIN.get(basename, []):
                    lines = re.sub(r"int main\(", "int {}_main(".format(
                        basename), lines)
                else:
                    lines = re.sub(r"int main\(", "int {}_{}_main(".format(
                        basename, subname), lines)
                lines = re.sub("stderr", "{}_stderr".format(basename), lines)
                lines = re.sub("stdout", "{}_stdout".format(basename), lines)
                lines = re.sub(r" printf\(", " fprintf({}_stdout, ".format(basename), lines)
                lines = re.sub(r"([^kf])puts\(", r"\1{}_puts(".format(basename), lines)
                lines = re.sub(r"putchar\(([^)]+)\)",
                               r"fputc(\1, {}_stdout)".format(basename), lines)

                fn = os.path.basename(filename)
                # some specific fixes:
                SPECIFIC_SUBSTITUTIONS = {
                    "bam_md.c": (
                        'sam_open_format("-", mode_w',
                        'sam_open_format({}_stdout_fn, mode_w'.format(basename)),
                    "phase.c": (
                        'putc("ACGT"[f->seq[j] == 1? (c&3, {}_stdout) : (c>>16&3)]);'.format(basename),
                        'putc("ACGT"[f->seq[j] == 1? (c&3) : (c>>16&3)], {}_stdout);'.format(basename)),
                    "cut_target.c": (
                        'putc(33 + (cns[j]>>8>>2, {}_stdout));'.format(basename),
                        'putc(33 + (cns[j]>>8>>2), {}_stdout);'.format(basename))
                    }
                if fn in SPECIFIC_SUBSTITUTIONS:
                    lines = lines.replace(
                        SPECIFIC_SUBSTITUTIONS[fn][0],
                        SPECIFIC_SUBSTITUTIONS[fn][1])
                outfile.write(lines)

    with open(os.path.join("import", "pysam.h")) as inf, \
         open(os.path.join(destdir, "{}.pysam.h".format(basename)), "w") as outf:
        outf.write(re.sub("@pysam@", basename, inf.read()))

    with open(os.path.join("import", "pysam.c")) as inf, \
         open(os.path.join(destdir, "{}.pysam.c".format(basename)), "w") as outf:
        outf.write(re.sub("@pysam@", basename, inf.read()))


if len(sys.argv) >= 1:
    if len(sys.argv) != 3:
        raise ValueError("import requires dest src")

    dest, srcdir = sys.argv[1:3]
    if dest not in EXCLUDE:
        raise ValueError("import expected one of %s" %
                         ",".join(EXCLUDE.keys()))
    exclude = EXCLUDE[dest]
    destdir = os.path.abspath(dest)
    srcdir = os.path.abspath(srcdir)
    if not os.path.exists(srcdir):
        raise IOError(
            "source directory `%s` does not exist." % srcdir)

    cfiles = locate("*.c", srcdir)
    hfiles = locate("*.h", srcdir)
    mfiles = itertools.chain(locate("README", srcdir), locate("LICENSE", srcdir))
    
    # remove unwanted files and htslib subdirectory.
    cfiles = [x for x in cfiles if os.path.basename(x) not in exclude
              and not re.search("htslib-", x)]

    hfiles = [x for x in hfiles if os.path.basename(x) not in exclude
              and not re.search("htslib-", x)]

    ncopied = 0

    def _compareAndCopy(src, srcdir, destdir, exclude):

        d, f = os.path.split(src)
        common_prefix = os.path.commonprefix((d, srcdir))
        subdir = re.sub(common_prefix, "", d)[1:]
        targetdir = os.path.join(destdir, subdir)
        if not os.path.exists(targetdir):
            os.makedirs(targetdir)
        old_file = os.path.join(targetdir, f)
        if os.path.exists(old_file):
            md5_old = hashlib.md5(
                "".join(open(old_file, "r", encoding="utf-8").readlines()).encode()).digest()
            md5_new = hashlib.md5(
                "".join(open(src, "r", encoding="utf-8").readlines()).encode()).digest()
            if md5_old != md5_new:
                raise ValueError(
                    "incompatible files for %s and %s" %
                    (old_file, src))

        shutil.copy(src, targetdir)
        return old_file

    for src_file in hfiles:
        _compareAndCopy(src_file, srcdir, destdir, exclude)
        ncopied += 1

    for src_file in mfiles:
        _compareAndCopy(src_file, srcdir, destdir, exclude)
        ncopied += 1

    cf = []
    for src_file in cfiles:
        cf.append(_compareAndCopy(src_file,
                                  srcdir,
                                  destdir,
                                  exclude))
        ncopied += 1

    sys.stdout.write(
        "installed latest source code from %s: "
        "%i files copied\n" % (srcdir, ncopied))
    # redirect stderr to pysamerr and replace bam.h with a stub.
    sys.stdout.write("applying stderr redirection\n")

    _update_pysam_files(cf, destdir)

    sys.exit(0)


# if len(sys.argv) >= 2 and sys.argv[1] == "refresh":
#     sys.stdout.write("refreshing latest source code from .c to .pysam.c")
#     # redirect stderr to pysamerr and replace bam.h with a stub.
#     sys.stdout.write("applying stderr redirection")
#     for destdir in ('samtools', ):
#         pysamcfiles = locate("*.pysam.c", destdir)
#         for f in pysamcfiles:
#             os.remove(f)
#         cfiles = locate("*.c", destdir)
#         _update_pysam_files(cfiles, destdir)

#     sys.exit(0)




import os
import subprocess
import pysam

try:
    import cyvcf2
except ImportError:
    pass
    

from TestUtils import CBCF_DATADIR, force_str

def build_filter_from_vcf_with_samtoolsshell(fn):
    retval = os.popen(
        "bcftools filter -e \"N_ALT != 1 || QUAL < 20 || maf[0]>0.05\" {} | grep -cv ^# ".format(fn)).read()
    return int(retval.strip())


def build_filter_from_vcf_with_bcftoolspipe(fn):
    FNULL = open(os.devnull, 'w')
    with subprocess.Popen([
            "bcftools",
            "filter",
            "-e",
            "N_ALT != 1 || QUAL < 20 || maf[0]>0.05",
            fn],
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=FNULL) as proc:
        data = [line for line in proc.stdout.readlines() if not line.startswith(b"#")]
        return len(data)


def build_filter_from_vcf_with_cyvcf2(fn):
    n = 0
    try:
        for v in cyvcf2.VCF(fn):
            if len(v.ALT) > 1:
                continue
            if v.QUAL < 20:
                continue
            if v.aaf > 0.05:
                continue
            n += 1
    except NameError:
        n = 9120
    return n


def build_filter_from_vcf_with_pysam(fn):
    n = 0
    with pysam.VariantFile(fn) as vcf:
        for v in vcf:
            # the two commands below take >1s out of 19s total
            if len(v.alts) > 1:
                continue
            if v.qual < 20:
                continue
            # this takes 12s out of 19s total
            gts = [s['GT'] for s in v.samples.values()]
            # the lines below take 6s out of 19s total
            an = sum(len(gt) for gt in gts)
            ac = sum(sum(gt) for gt in gts)
            aaf = (float(ac) / float(an))
            if aaf > 0.05:
                continue
            n += 1
    return n
                                        



import os
import glob
import sys
import unittest
import pysam
import shutil
import gzip
import subprocess
import pytest

try:
    from pathlib import Path
except ImportError:
    Path = None

from TestUtils import get_temp_filename, check_lines_equal, load_and_convert, CBCF_DATADIR, get_temp_context


@pytest.fixture
def vcf_header():
    vcf_header = pysam.VariantHeader()
    vcf_header.add_sample("sample1")
    vcf_header.add_sample("sample2")
    vcf_header.contigs.add("1")
    return vcf_header

## segfault without coordinates

def test_ascii_annotation_can_be_added(vcf_header):
    vcf_header.formats.add("AN", 1, "String", "An annotation")
    record = vcf_header.new_record(
        contig="1",
        start=12,
        stop=13,
        samples=[
            {"AN": "anno1"},
            {"AN": "anno2"}])
    assert str(record)[:-1].split("\t")[-2:] == ["anno1", "anno2"]


def test_ascii_annotation_with_variable_length_can_be_added(vcf_header):
    vcf_header.formats.add("AN", 1, "String", "An annotation")
    record = vcf_header.new_record(
        contig="1",
        start=12,
        stop=13,
        samples=[
            {"AN": "anno1b"},
            {"AN": "anno1"}])
    assert str(record)[:-1].split("\t")[-2:] == ["anno1b", "anno1"]
    record = vcf_header.new_record(
        contig="1",
        start=12,
        stop=13,
        samples=[
            {"AN": "anno2"},
            {"AN": "anno2b"}])
    assert str(record)[:-1].split("\t")[-2:] == ["anno2", "anno2b"]
    

def test_unicode_annotation_can_be_added(vcf_header):
    vcf_header.formats.add("AN", 1, "String", "An annotation")
    record = vcf_header.new_record(
        contig="1",
        start=12,
        stop=13,
        samples=[
            {"AN": "anno1"},
            {"AN": "Friedrich-Alexander-Universit\u00E4t_Erlangen-N\u00FCrnberg"}])
    assert str(record)[:-1].split("\t")[-2:] == [
        "anno1",
        "Friedrich-Alexander-Universit\u00E4t_Erlangen-N\u00FCrnberg"]



"""Benchmarking module for AlignmentFile functionality"""
import os
import pytest


from TestUtils import BAM_DATADIR, force_str, flatten_nested_list
from AlignmentFileFetchTestUtils import *


def test_build_fetch_from_bam_with_samtoolsshell(benchmark):
    result = benchmark(build_fetch_with_samtoolsshell,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 3270


def test_build_fetch_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_fetch_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 3270


def test_build_fetch_from_bam_with_pysam(benchmark):
    result = benchmark(build_fetch_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 3270


def test_build_query_sequences_from_bam_with_samtoolsshell(benchmark):
    result = benchmark(build_query_sequences_with_samtoolsshell,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3270


def test_build_query_sequences_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_query_sequences_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3270


def test_build_query_sequences_from_bam_with_pysam(benchmark):
    result = benchmark(build_query_sequences_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3270


def test_build_query_qualities_from_bam_with_pysam(benchmark):
    result = benchmark(build_query_qualities_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3270


def test_build_query_sequences_from_bam_flagfilter_with_samtoolsshell(benchmark):
    result = benchmark(build_query_sequences_flagfilter_with_samtoolsshell,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3124


def test_build_query_sequences_from_bam_flagfilter_with_samtoolspipe(benchmark):
    result = benchmark(build_query_sequences_flagfilter_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3124


def test_build_query_sequences_from_bam_flagfilter_with_pysam(benchmark):
    result = benchmark(build_query_sequences_flagfilter_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3124


def test_build_query_sequences_from_bam_directflagfilter_with_pysam(benchmark):
    result = benchmark(build_query_sequences_flagfilter_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len(result) == 3124


@pytest.mark.aligned_pairs
def test_build_aligned_pairs_default_with_pysam(benchmark):
    result = benchmark(build_aligned_pairs_with_pysam,
                       os.path.join(BAM_DATADIR, "with_md.bam"))
    assert len(result) == 3235


@pytest.mark.aligned_pairs
def test_build_aligned_pairs_matchesonly_with_pysam(benchmark):
    result = benchmark(build_aligned_pairs_with_pysam,
                       os.path.join(BAM_DATADIR, "with_md.bam"),
                       matches_only=True)
    assert len(result) == 3235


@pytest.mark.aligned_pairs    
def test_build_aligned_pairs_withseq_with_pysam(benchmark):
    result = benchmark(build_aligned_pairs_with_pysam,
                       os.path.join(BAM_DATADIR, "with_md.bam"),
                       with_seq=True)
    assert len(result) == 3235
    




"""test linking against pysam.
"""

import unittest
import os
import subprocess
import pysam

from TestUtils import LINKDIR


def check_import(statement):
    try:
        output = subprocess.check_output(
            statement, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as exc:
        if b"ImportError" in exc.output:
            raise ImportError(
                "module could not be imported: {}".format(str(exc.output)))
        else:
            raise


def check_pass(statement):
    try:
        output = subprocess.check_output(
            statement, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as exc:
        raise ValueError("{}: {}".format(exc, exc.output))
    if b"FAILED" in output:
        raise ValueError("module tests failed")
    return True


@unittest.skipUnless(
    os.environ.get("PYSAM_LINKING_TESTS", None),
    "enable linking tests by setting PYSAM_LINKING_TESTS environment variable")
class TestLinking(unittest.TestCase):

    package_name = "link_with_rpath"

    def setUp(self):
        self.workdir = os.path.join(LINKDIR, self.package_name)

    def test_package_can_be_installed(self):
        subprocess.check_output(
            "cd {} && rm -rf build && python setup.py install".format(
                self.workdir),
            shell=True)


@unittest.skipUnless(
    os.environ.get("PYSAM_LINKING_TESTS", None),
    "enable linking tests by setting PYSAM_LINKING_TESTS environment variable")
class TestLinkWithRpath(TestLinking):

    package_name = "link_with_rpath"

    def test_package_tests_pass(self):
        self.assertTrue(check_pass(
            "cd {} && python test_module.py".format(os.path.join(self.workdir, "tests"))))


@unittest.skipUnless(
    os.environ.get("PYSAM_LINKING_TESTS", None),
    "enable linking tests by setting PYSAM_LINKING_TESTS environment variable")
class TestLinkWithoutRpath(TestLinking):

    package_name = "link_without_rpath"

    def test_package_tests_fail_on_import(self):

        self.assertRaises(
            ImportError,
            check_import,
            "cd {} && python test_module.py".format(os.path.join(self.workdir, "tests")))

    def test_package_tests_pass_if_ld_library_path_set(self):

        pysam_libraries = pysam.get_libraries()
        pysam_libdirs, pysam_libs = zip(
            *[os.path.split(x) for x in pysam_libraries])
        pysam_libdir = pysam_libdirs[0]

        self.assertTrue(check_pass(
            "export LD_LIBRARY_PATH={}:$PATH && cd {} && python test_module.py".format(
                pysam_libdir,
                os.path.join(self.workdir, "tests"))))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
'''unit testing code for pysam.

Execute in the :file:`tests` directory as it requires the Makefile
and data files located there.
'''

import warnings
import unittest
import os
import re
import glob
import sys
import subprocess
import shutil
import pysam
import pysam.samtools
import pysam.bcftools
from TestUtils import checkBinaryEqual, check_lines_equal, \
    check_samtools_view_equal, get_temp_filename, force_bytes, WORKDIR, \
    BAM_DATADIR


IS_PYTHON3 = sys.version_info[0] >= 3


def run_command(cmd):
    '''run a samtools command'''
    try:
        retcode = subprocess.call(cmd, shell=True,
                                  stderr=subprocess.PIPE)
        if retcode < 0:
            print("Child was terminated by signal", -retcode)
    except OSError as e:
        print("Execution failed:", e)


def get_version(executable):
    '''return samtools/bcftools version'''

    with subprocess.Popen(executable, shell=True,
                          stderr=subprocess.PIPE).stderr as pipe:
        lines = b"".join(pipe.readlines())

    if IS_PYTHON3:
        lines = lines.decode('ascii')
    try:
        x = re.search(r"Version:\s+(\S+)", lines).groups()[0]
    except AttributeError:
        raise ValueError("could not get version from %s" % lines)
    return x


class SamtoolsTest(unittest.TestCase):

    '''test samtools command line commands and compare
    against pysam commands.

    Tests fail, if the output is not binary identical.
    '''

    requisites = [
        "ex1.fa", "ex1.fa.fai",
        "ex1.sam.gz",
        "ex1.bam", "ex1.bam.bai",
        "ex1.sam",
        "ex1.sam",
        "ex2.bam",
        "ex2.sam",
        "ex1.bed"]

    # a list of statements to test
    # should contain at least one %(out)s component indicating
    # an output file.
    statements = [
        "view ex1.bam > %(out)s_ex1.view",
        "view -c ex1.bam > %(out)s_ex1.count",
        # ("view -bT ex1.fa -o %(out)s_ex1.view2 ex1.sam",
        "sort ex1.bam -o %(out)s_ex1.sort.bam",
        "mpileup ex1.bam > %(out)s_ex1.pileup",
        "depth ex1.bam > %(out)s_ex1.depth",
        # TODO: issues with file naming
        # "faidx ex1.fa; %(out)s_ex1.fa.fai",
        "index ex1.bam %(out)s_ex1.bam.fai",
        "index -@2 ex1.bam %(out)s_ex1.bam.fai",
        "idxstats ex1.bam > %(out)s_ex1.idxstats",
        "fixmate ex1.bam %(out)s_ex1.fixmate.bam",
        "flagstat ex1.bam > %(out)s_ex1.flagstat",
        # Fails python 3.3 on linux, passes on OsX and when
        # run locally
        "calmd ex1.bam ex1.fa > %(out)s_ex1.calmd.bam",
        # use -s option, otherwise the following error in samtools 1.2:
        # Samtools-htslib-API: bam_get_library() not yet implemented
        # causes downstream problems
        # TODO: The following cause subsequent commands to fail
        # unknow option
        # "rmdup -s ex1.bam %(out)s_ex1.rmdup.bam",
        # "merge -f %(out)s_ex1.merge.bam ex1.bam ex1.bam",
        "reheader ex2.sam ex1.bam > %(out)s_ex1.reheader.bam",
        "cat -o %(out)s_ex1.cat.bam ex1.bam ex1.bam",
        "targetcut ex1.bam > %(out)s_ex1.targetcut",
        "phase ex1.bam > %(out)s_ex1.phase",
        "import ex1.fa.fai ex1.sam.gz %(out)s_ex1.bam",
        "bam2fq ex1.bam > %(out)s_ex1.bam2fq",
        # TODO: not the same
        # "pad2unpad -T ex1.fa ex2.bam > %(out)s_ex2.unpad",
        # TODO: command line option problem
        # "bamshuf ex1.bam -O --output-fmt SAM > %(out)s_ex1.bamshuf.sam",
        # "collate ex1.bam %(out)s_ex1.collate",
        "bedcov ex1.bed ex1.bam > %(out)s_ex1.bedcov",
        "stats ex1.bam > %(out)s_ex1.stats",
        "dict ex1.bam > %(out)s_ex1.dict",
        # TODO: not the same
        # ("addreplacerg -r 'RG\tID:ga\tSM:hs' ex1.bam > %(out)s_ex1.addreplacerg",
    ]

    map_command = {
        "import": "samimport"}

    executable = "samtools"

    module = pysam.samtools

    def check_version(self):

        samtools_version = get_version(self.executable)

        def _r(s):
            # patch - remove any of the alpha/beta suffixes, i.e., 0.1.12a ->
            # 0.1.12
            if s.count('-') > 0:
                s = s[0:s.find('-')]
            return re.sub("[^0-9.]", "", s)

        if _r(samtools_version) != _r(pysam.__samtools_version__):
            warnings.warn(
                "versions of pysam.%s and %s differ: %s != %s" %
                (self.executable,
                 self.executable,
                 pysam.__samtools_version__,
                 samtools_version))

    def setUp(self):
        '''setup tests.

        For setup, all commands will be run before the first test is
        executed. Individual tests will then just compare the output
        files.

        '''
        self.check_version()

        self.workdir = os.path.join(WORKDIR, "samtools_test")
        
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

        for f in self.requisites:
            shutil.copy(os.path.join(BAM_DATADIR, f),
                        os.path.join(self.workdir, f))

        self.savedir = os.getcwd()
        os.chdir(self.workdir)

        return

    def get_command(self, statement, map_to_internal=True):
        """return samtools command from statement"""
        parts = statement.split(" ")
        command = parts[0]
        if map_to_internal:
            return self.map_command.get(command, command)
        else:
            return command

    def check_statement(self, statement):

        parts = statement.split(" ")
        r_samtools = {"out": self.executable}
        r_pysam = {"out": "pysam"}

        command = self.get_command(statement)

        # self.assertTrue(command in pysam.SAMTOOLS_DISPATCH)

        targets = [x for x in parts if "%(out)s" in x]
        samtools_targets = [x % r_samtools for x in targets]
        pysam_targets = [x % r_pysam for x in targets]

        pysam_method = getattr(self.module, command)

        # run samtools
        full_statement = re.sub(r"%\(out\)s", self.executable, statement)
        run_command(" ".join((self.executable, full_statement)))
        # sys.stdout.write("%s %s ok" % (command, self.executable))

        # run pysam
        if ">" in statement:
            assert parts[-2] == ">"
            parts = parts[:-2]

        # avoid interpolation to preserve string quoting, tab chars, etc.
        pysam_parts = [re.sub(r"%\(out\)s", "pysam", x) for x in parts[1:]]
        output = pysam_method(*pysam_parts,
                              raw=True,
                              catch_stdout=True)
        # sys.stdout.write(" pysam ok\n")
        if ">" in statement:
            with open(pysam_targets[-1], "wb") as outfile:
                if output is not None:
                    outfile.write(force_bytes(output))
        for samtools_target, pysam_target in zip(samtools_targets,
                                                 pysam_targets):
            if os.path.isdir(samtools_target):
                samtools_files = glob.glob(os.path.join(
                    samtools_target, "*"))
                pysam_files = glob.glob(os.path.join(pysam_target, "*"))
                self.assertEqual(len(samtools_files), len(pysam_files))
                # need to be able to exclude files like README, etc.
                continue
            else:
                samtools_files = [samtools_target]
                pysam_files = [pysam_target]

            for s, p in zip(samtools_files, pysam_files):
                binary_equal = checkBinaryEqual(s, p)
                error_msg = "%s failed: files %s and %s are not the same" % (
                    command, s, p)
                if binary_equal:
                    continue
                elif s.endswith(".bam"):
                    self.assertTrue(
                        check_samtools_view_equal(
                            s, p, without_header=True),
                        error_msg)
                else:
                    check_lines_equal(
                        self, s, p,
                        filter_f=lambda x: x.startswith("#"),
                        msg=error_msg)

    def testStatements(self):
        for statement in self.statements:
            command = self.get_command(statement, map_to_internal=False)
            # bam2fq differs between version 1.5 and 1.6 - reenable if
            # bioconda samtools will be available.
            if command in ("bedcov", "stats", "dict", "bam2fq"):
                continue

            if (command == "calmd" and
                    list(sys.version_info[:2]) == [3, 3]):
                # skip calmd test, fails only on python 3.3.5
                # in linux (empty output). Works in OsX and passes
                # for 3.4 and 3.5, see issue #293
                continue
            self.check_statement(statement)

    @unittest.skipIf(sys.platform == "darwin", "not supported, pattern does not match")
    @unittest.skipIf(not sys.stdin.isatty(), "skipping usage tests, stdin is not a tty")
    def testUsage(self):
        if self.executable == "bcftools":
            # bcftools usage messages end with exit(1)
            return

        for statement in self.statements:
            command = self.get_command(statement, map_to_internal=False)
            # ignore commands that exit or cause other failures
            # TODO: check - if reheader or phase is run in testStatements, sort fails
            # here
            if command in ("view", "sort", "bam2fq", "flagstat", "reheader",
                           "stats", "idxstats"):
                continue
            mapped_command = self.get_command(statement, map_to_internal=True)
            pysam_method = getattr(self.module, mapped_command)
            usage_msg = pysam_method.usage()
            expected = r"Usage:\s+{} {}".format(self.executable, command)
            self.assertTrue(re.search(expected, usage_msg) is not None)

    def tearDown(self):
        if os.path.exists(self.workdir):
            shutil.rmtree(self.workdir)
        os.chdir(self.savedir)


class EmptyIndexTest(unittest.TestCase):

    def testEmptyIndex(self):
        self.assertRaises(IOError, pysam.samtools.index,
                          "exdoesntexist.bam")

    def testEmptyIndexWithExtraFlag(self):
        self.assertRaises(IOError, pysam.samtools.index,
                          "-c",
                          "exdoesntexist.bam")

    def testEmptyIndexWithExtraArg(self):
        self.assertRaises(IOError, pysam.samtools.index,
                          "-c",
                          "-m",
                          "14",
                          "exdoesntexist.bam")


if sys.platform != "darwin":
    # fails with segfault with htslib 1.5 on Osx, an issue with flockfile
    # issue seems to be with repeated calls to interface

    class TestReturnType(unittest.TestCase):

        def testReturnValueString(self):
            retval = pysam.idxstats(os.path.join(BAM_DATADIR, "ex1.bam"))
            if IS_PYTHON3:
                self.assertFalse(isinstance(retval, bytes))
                self.assertTrue(isinstance(retval, str))
            else:
                self.assertTrue(isinstance(retval, bytes))
                self.assertTrue(isinstance(retval, basestring))

        def testReturnValueData(self):
            args = "-O BAM {}".format(os.path.join(BAM_DATADIR,
                                                   "ex1.bam")).split(" ")
            retval = pysam.view(*args)

            if IS_PYTHON3:
                self.assertTrue(isinstance(retval, bytes))
                self.assertFalse(isinstance(retval, str))
            else:
                self.assertTrue(isinstance(retval, bytes))
                self.assertTrue(isinstance(retval, basestring))

    class StdoutTest(unittest.TestCase):
        '''test if stdout can be redirected.'''

        def testWithRedirectedStdout(self):
            r = pysam.samtools.flagstat(
                os.path.join(BAM_DATADIR, "ex1.bam"))
            self.assertTrue(len(r) > 0)

        def testWithoutRedirectedStdout(self):
            r = pysam.samtools.flagstat(
                os.path.join(BAM_DATADIR, "ex1.bam"),
                catch_stdout=False)
            self.assertEqual(r, None)

        def testDoubleCalling(self):
            # The following would fail if there is an
            # issue with stdout being improperly caught.
            retvals = pysam.idxstats(
                os.path.join(BAM_DATADIR, "ex1.bam"))
            retvals = pysam.idxstats(
                os.path.join(BAM_DATADIR, "ex1.bam"))

        def testSaveStdout(self):
            outfile = get_temp_filename(suffix=".tsv")
            r = pysam.samtools.flagstat(
                os.path.join(BAM_DATADIR, "ex1.bam"),
                save_stdout=outfile)
            self.assertEqual(r, None)
            with open(outfile) as inf:
                r = inf.read()
            self.assertTrue(len(r) > 0)

    class PysamTest(SamtoolsTest):
        """check access to samtools command in the pysam
        main package.

        This is for backwards capability.
        """

        module = pysam

# class BcftoolsTest(SamtoolsTest):

#     requisites = [
#         "ex1.fa",
#         "ex1.vcf.gz",
#         "ex1.vcf.gz.tbi",
#     ]
#     # a list of statements to test
#     # should contain at least one %(out)s component indicating
#     # an output file.
#     statements = [
#         # "index -n ex1.vcf.gz > %(out)s_ex1.index",

#         "annotate -x ID ex1.vcf.gz > %(out)s_ex1.annotate",
#         "concat -a ex1.vcf.gz ex1.vcf.gz > %(out)s_ex1.concat",
#         "isec -p %(out)s_ex1.isec ex1.vcf.gz ex1.vcf.gz",
#         "merge --force-samples ex1.vcf.gz ex1.vcf.gz > %(out)s_ex1.norm",
#         "norm -m +both ex1.vcf.gz > %(out)s_ex1.norm",

#         # "plugin",
#         # "query -f '%CHROM\n' ex1.vcf.gz > %(out)s_ex1.query",
#         # "reheader -s A > %(out)s_ex1.reheader",
#         # "view ex1.vcf.gz > %(out)s_ex1.view",
#         # "call -m ex1.vcf.gz > %(out)s_ex1.call",
#         # bad file descriptor
#         # "consensus -f ex1.fa ex1.vcf.gz  > %(out)s_ex1.consensus"
#         # need appropriate VCF file
#         # "cnv",
#         # segfault
#         # "filter -s A ex1.vcf.gz  > %(out)s_ex1.filter",
#         # exit
#         # "gtcheck -s A ex1.vcf.gz  > %(out)s_ex1.gtcheck",
#         # segfauld, used to work wit bcftools 1.3
#         # "roh -s A ex1.vcf.gz > %(out)s_ex1.roh",
#         "stats ex1.vcf.gz > %(out)s_ex1.stats",
#     ]

#     map_command = {
#         "import": "samimport"}

#     executable = "bcftools"

#     module = pysam.bcftools


if __name__ == "__main__":
    # build data files
    print("building data files")
    subprocess.call("make -C %s" % BAM_DATADIR, shell=True)
    print("starting tests")
    unittest.main()
    print("completed tests")



import pysam
import os
from TestUtils import BAM_DATADIR


def test_idxstats_parse_split_lines():
    bam_filename = os.path.join(BAM_DATADIR, "ex2.bam")
    # Test pysam 0.8.X style output, which returns a list of lines
    lines = pysam.idxstats(bam_filename, split_lines=True)
    for line in lines:
        _seqname, _seqlen, nmapped, _nunmapped = line.split()


def test_bedcov_split_lines():
    bam_filename = os.path.join(BAM_DATADIR, "ex1.bam")
    bed_filename = os.path.join(BAM_DATADIR, "ex1.bed")
    # Test pysam 0.8.X style output, which returns a list of lines
    lines = pysam.bedcov(bed_filename, bam_filename, split_lines=True)
    for line in lines:
        fields = line.split('\t')
        assert len(fields) in [4, 5], \
            ("bedcov should give tab delimited output with 4 or 5 fields. "
             "Split line (%s) gives %d fields." % (fields, len(fields)))


def test_idxstats_parse():
    bam_filename = os.path.join(BAM_DATADIR, "ex2.bam")
    # Test pysam 0.9.X style output, which returns a string that needs to be split by \n
    idxstats_string = pysam.idxstats(bam_filename, split_lines=False)
    lines = idxstats_string.splitlines()
    for line in lines:
        splt = line.split("\t")
        _seqname, _seqlen, nmapped, _nunmapped = splt


def test_bedcov():
    bam_filename = os.path.join(BAM_DATADIR, "ex1.bam")
    bed_filename = os.path.join(BAM_DATADIR, "ex1.bed")
    # Test pysam 0.9.X style output, which returns a string that needs to be split by \n
    bedcov_string = pysam.bedcov(bed_filename, bam_filename, split_lines=False)
    lines = bedcov_string.splitlines()
    for line in lines:
        fields = line.split('\t')
        assert len(fields) in [4, 5], \
            ("bedcov should give tab delimited output with 4 or 5 fields. "
             "Split line (%s) gives %d fields." % (fields, len(fields)))



import gzip
import os
import pysam

from TestUtils import TABIX_DATADIR

FN_COMPRESSED = "example.bed.gz"
FN_UNCOMPRESSED = "example.bed"
FN_LARGE_COMPRESSED = "example_large.bed.gz"
FN_LARGE_UNCOMPRESSED = "example_large.bed"


def read_python_compressed(fn):
    '''iterate through with python.'''
    with gzip.open(fn, mode="r") as f:
        return len([x.split(b"\t") for x in f])


def read_python_uncompressed(fn):
    with open(fn) as f:
        return len([x.split("\t") for x in f])


def fetch_plain(fn):
    with pysam.Tabixfile(fn) as f:
        return len(list(f.fetch()))


def fetch_parsed(fn):
    with pysam.Tabixfile(fn) as f:
        return len(list(f.fetch(parser=pysam.asBed())))


def iterate_generic_compressed(fn):
    with gzip.open(fn) as f:
        return len(list(pysam.tabix_generic_iterator(f, parser=pysam.asBed())))


def iterate_generic_uncompressed(fn):
    with open(fn) as f:
        return len(list(pysam.tabix_generic_iterator(f, parser=pysam.asBed())))


def iterate_parsed_compressed(fn):
    with gzip.open(fn) as f:
        return len(list(pysam.tabix_iterator(f, parser=pysam.asBed())))


def iterate_parsed_uncompressed(fn):
    with open(fn) as f:
        return len(list(pysam.tabix_iterator(f, parser=pysam.asBed())))


def iterate_file_compressed(fn):
    with gzip.open(fn) as f:
        return len(list(pysam.tabix_file_iterator(f, parser=pysam.asBed())))


def iterate_file_uncompressed(fn):
    with open(fn) as f:
        return len(list(pysam.tabix_file_iterator(f, parser=pysam.asBed())))


def test_read_python_compressed(benchmark):
    result = benchmark(read_python_compressed,
                       os.path.join(TABIX_DATADIR, FN_COMPRESSED))
    assert result == 164


def test_read_python_uncompressed(benchmark):
    result = benchmark(read_python_uncompressed,
                       os.path.join(TABIX_DATADIR, FN_UNCOMPRESSED))
    assert result == 164


def test_fetch_plain(benchmark):
    result = benchmark(fetch_plain, os.path.join(TABIX_DATADIR, FN_COMPRESSED))
    assert result == 164


def test_fetch_parsed(benchmark):
    result = benchmark(fetch_parsed, os.path.join(
        TABIX_DATADIR, FN_COMPRESSED))
    assert result == 164


def test_iterate_generic_compressed(benchmark):
    result = benchmark(iterate_generic_compressed,
                       os.path.join(TABIX_DATADIR, FN_COMPRESSED))
    assert result == 164


def test_iterate_generic_uncompressed(benchmark):
    result = benchmark(iterate_generic_uncompressed,
                       os.path.join(TABIX_DATADIR, FN_UNCOMPRESSED))
    assert result == 164


def test_iterate_parsed_compressed(benchmark):
    result = benchmark(iterate_parsed_compressed,
                       os.path.join(TABIX_DATADIR, FN_COMPRESSED))
    assert result == 164


def test_iterate_parsed_uncompressed(benchmark):
    result = benchmark(iterate_parsed_uncompressed,
                       os.path.join(TABIX_DATADIR, FN_UNCOMPRESSED))
    assert result == 164


def test_iterate_file_compressed(benchmark):
    result = benchmark(iterate_file_compressed,
                       os.path.join(TABIX_DATADIR, FN_COMPRESSED))
    assert result == 164


def test_iterate_file_uncompressed(benchmark):
    result = benchmark(iterate_file_uncompressed,
                       os.path.join(TABIX_DATADIR, FN_UNCOMPRESSED))
    assert result == 164


def test_read_python_large_compressed(benchmark):
    result = benchmark(read_python_compressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_COMPRESSED))
    assert result == 100000


def test_read_python_large_uncompressed(benchmark):
    result = benchmark(read_python_uncompressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_UNCOMPRESSED))
    assert result == 100000


def test_fetch_plain(benchmark):
    result = benchmark(fetch_plain, os.path.join(
        TABIX_DATADIR, FN_LARGE_COMPRESSED))
    assert result == 100000


def test_fetch_parsed(benchmark):
    result = benchmark(fetch_parsed, os.path.join(
        TABIX_DATADIR, FN_LARGE_COMPRESSED))
    assert result == 100000


def test_iterate_generic_large_compressed(benchmark):
    result = benchmark(iterate_generic_compressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_COMPRESSED))
    assert result == 100000


def test_iterate_generic_large_uncompressed(benchmark):
    result = benchmark(iterate_generic_uncompressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_UNCOMPRESSED))
    assert result == 100000


def test_iterate_parsed_large_compressed(benchmark):
    result = benchmark(iterate_parsed_compressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_COMPRESSED))
    assert result == 100000


def test_iterate_parsed_large_uncompressed(benchmark):
    result = benchmark(iterate_parsed_uncompressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_UNCOMPRESSED))
    assert result == 100000


def test_iterate_file_large_compressed(benchmark):
    result = benchmark(iterate_file_compressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_COMPRESSED))
    assert result == 100000


def test_iterate_file_large_uncompressed(benchmark):
    result = benchmark(iterate_file_uncompressed, os.path.join(
        TABIX_DATADIR, FN_LARGE_UNCOMPRESSED))
    assert result == 100000



#!/usr/bin/env python
'''unit testing code for pysam.

Execute in the :file:`tests` directory as it requires the Makefile
and data files located there.
'''

import unittest
import os
import sys
import re
import copy
from collections import OrderedDict as odict
import pysam
import pysam.samtools
from TestUtils import get_temp_filename, BAM_DATADIR

if sys.version_info.major >= 3:
    from io import StringIO
else:
    from StringIO import StringIO


class TestHeaderConstruction(unittest.TestCase):
    """testing header construction."""

    header_dict = odict(
        [('SQ', [odict([('LN', 1575), ('SN', 'chr1'), ('AH', 'chr1:5000000-5010000')]),
                 odict([('LN', 1584), ('SN', 'chr2'), ('AH', '*')])]),
         ('RG', [odict([('LB', 'SC_1'), ('ID', 'L1'), ('SM', 'NA12891'),
                        ('PU', 'SC_1_10'), ("CN", "name:with:colon")]),
                 odict([('LB', 'SC_2'), ('ID', 'L2'), ('SM', 'NA12891'),
                        ('PU', 'SC_2_12'), ("CN", "name:with:colon")])]),
         ('PG', [odict([('ID', 'P1'), ('VN', '1.0')]),
                 odict([('ID', 'P2'), ('VN', '1.1')])]),
         ('HD', odict([('VN', '1.0')])),
         ('CO', ['this is a comment', 'this is another comment']),
        ])

    header_text = ("@HD\tVN:1.0\n"
                   "@SQ\tSN:chr1\tLN:1575\tAH:chr1:5000000-5010000\n"
                   "@SQ\tSN:chr2\tLN:1584\tAH:*\n"
                   "@RG\tID:L1\tPU:SC_1_10\tLB:SC_1\tSM:NA12891\tCN:name:with:colon\n"
                   "@RG\tID:L2\tPU:SC_2_12\tLB:SC_2\tSM:NA12891\tCN:name:with:colon\n"
                   "@PG\tID:P1\tVN:1.0\n"
                   "@PG\tID:P2\tVN:1.1\n"
                   "@CO\tthis is a comment\n"
                   "@CO\tthis is another comment\n")


    header_from_references = odict(
        [('SQ', [odict([('LN', 1575), ('SN', 'chr1')]),
                 odict([('LN', 1584), ('SN', 'chr2')])]),
         ('RG', [odict([('LB', 'SC_1'), ('ID', 'L1'), ('SM', 'NA12891'),
                        ('PU', 'SC_1_10'), ("CN", "name:with:colon")]),
                 odict([('LB', 'SC_2'), ('ID', 'L2'), ('SM', 'NA12891'),
                        ('PU', 'SC_2_12'), ("CN", "name:with:colon")])]),
         ('PG', [odict([('ID', 'P1'), ('VN', '1.0')]),
                 odict([('ID', 'P2'), ('VN', '1.1')])]),
         ('HD', odict([('VN', '1.0')])),
         ('CO', ['this is a comment', 'this is another comment']),
        ])

    header_without_text = odict(
        [('SQ', [odict([('LN', 1575), ('SN', 'chr1')]),
                 odict([('LN', 1584), ('SN', 'chr2')])]),
        ])
    
    def compare_headers(self, test_header, ref_header=None):
        '''compare two headers a and b.'''
        test_header_dict = test_header.as_dict()
        if ref_header is None:
            ref_header = self.header_dict
            
        for ak, av in test_header_dict.items():
            self.assertTrue(ak in self.header_dict, "key '%s' not in '%s' " % (ak, ref_header))
            self.assertEqual(av, ref_header[ak])
        for ak, av in ref_header.items():
            self.assertTrue(ak in test_header_dict, "key '%s' not in '%s' " % (ak, test_header_dict))
            self.assertEqual(av, test_header_dict[ak])

    def check_name_mapping(self, test_header):
        for x, y in enumerate(("chr1", "chr2")):
            tid = test_header.get_tid(y)
            ref = test_header.get_reference_name(x)
            self.assertEqual(tid, x)
            self.assertEqual(ref, y)

        self.assertEqual(test_header.get_tid("chr?"), -1)
        self.assertRaises(ValueError, test_header.get_reference_name, 2)
            
    def test_header_constructed_from_dict(self):
        header = pysam.AlignmentHeader.from_dict(self.header_dict)
        self.compare_headers(header)
        self.check_name_mapping(header)
        
    def test_header_constructed_from_text(self):
        header = pysam.AlignmentHeader.from_text(self.header_text)
        self.compare_headers(header)
        self.check_name_mapping(header)
        
    def test_header_constructed_from_header(self):
        header = pysam.AlignmentHeader.from_text(self.header_text)
        self.compare_headers(header.copy())
        self.check_name_mapping(header)

    def test_header_constructed_from_references(self):
        text = re.sub("@SQ[^\n]+\n", "", self.header_text)
        assert "@SQ" not in text
        header = pysam.AlignmentHeader.from_references(
            reference_names=["chr1", "chr2"],
            reference_lengths=[1575, 1584],
            text=text)
        self.compare_headers(header, self.header_from_references)
        self.check_name_mapping(header)

    def test_header_constructed_from_references_without_text(self):
        header = pysam.AlignmentHeader.from_references(
            reference_names=["chr1", "chr2"],
            reference_lengths=[1575, 1584])
        self.compare_headers(header, self.header_without_text)
        self.check_name_mapping(header)


class TestHeaderSAM(unittest.TestCase):
    """testing header manipulation"""

    header = {'SQ': [{'LN': 1575, 'SN': 'chr1', 'AH': 'chr1:5000000-5010000'},
                     {'LN': 1584, 'SN': 'chr2', 'AH': '*'}],
              'RG': [{'LB': 'SC_1', 'ID': 'L1', 'SM': 'NA12891',
                      'PU': 'SC_1_10', "CN": "name:with:colon"},
                     {'LB': 'SC_2', 'ID': 'L2', 'SM': 'NA12891',
                      'PU': 'SC_2_12', "CN": "name:with:colon"}],
              'PG': [{'ID': 'P1', 'VN': '1.0'}, {'ID': 'P2', 'VN': '1.1'}],
              'HD': {'VN': '1.0'},
              'CO': ['this is a comment', 'this is another comment'],
              }

    def compare_headers(self, a, b):
        '''compare two headers a and b.'''
        for ak, av in a.items():
            self.assertTrue(ak in b, "key '%s' not in '%s' " % (ak, b))
            self.assertEqual(av, b[ak])

    def setUp(self):
        self.samfile = pysam.AlignmentFile(
            os.path.join(BAM_DATADIR, "ex3.sam"),
            "r")

    def test_header_content_is_as_expected(self):
        self.compare_headers(self.header, self.samfile.header.to_dict())
        self.compare_headers(self.samfile.header.to_dict(), self.header)

    def test_text_access_works(self):
        self.assertEqual(self.samfile.text, self.samfile.header.__str__())
        
    def test_name_mapping(self):
        for x, y in enumerate(("chr1", "chr2")):
            tid = self.samfile.gettid(y)
            ref = self.samfile.getrname(x)
            self.assertEqual(tid, x)
            self.assertEqual(ref, y)

        self.assertEqual(self.samfile.gettid("chr?"), -1)
        self.assertRaises(ValueError, self.samfile.getrname, 2)

    def test_dictionary_access_works(self):
        for key in self.header.keys():
            self.compare_headers({key: self.header[key]},
                                 {key: self.samfile.header[key]})

    def test_dictionary_setting_raises_error(self):
        self.assertRaises(TypeError,
                          self.samfile.header.__setitem__,
                          "CO",
                          ["This is a final comment"])

    def test_dictionary_len_works(self):
        self.assertEqual(len(self.header), len(self.samfile.header))

    def test_dictionary_keys_works(self):
        # sort for py2.7
        self.assertEqual(sorted(self.header.keys()),
                         sorted(self.samfile.header.keys()))

    def test_dictionary_values_works(self):
        self.assertEqual(len(self.header.values()), len(self.samfile.header.values()))
        
    def test_dictionary_get_works(self):
        self.assertEqual(self.header.get("HD"), {'VN': '1.0'})
        self.assertEqual(self.header.get("UK", "xyz"), "xyz")
        self.assertEqual(self.header.get("UK"), None)

    def test_dictionary_contains_works(self):
        self.assertTrue("HD" in self.header)
        self.assertFalse("UK" in self.header)

    def tearDown(self):
        self.samfile.close()


class TestHeaderBAM(TestHeaderSAM):

    def setUp(self):
        self.samfile = pysam.AlignmentFile(
            os.path.join(BAM_DATADIR, "ex3.bam"),
            "rb")


class TestHeaderCRAM(TestHeaderSAM):

    def setUp(self):
        self.samfile = pysam.AlignmentFile(
            os.path.join(BAM_DATADIR, "ex3.cram"),
            "rc")

    def compare_headers(self, a, b):
        '''compare two headers a and b.'''
        def _strip(dd):
            for x in dd:
                for y in ("M5", "UR"):
                    if y in x:
                        del x[y]
        for ak, av in a.items():
            _strip(av)
            self.assertTrue(ak in b, "key '%s' not in '%s' " % (ak, b))
            _strip(b[ak])

            self.assertEqual(av, b[ak])


class TestHeaderFromRefs(unittest.TestCase):
    '''see issue 144

    reference names need to be converted to string for python 3
    '''

    # def testHeader( self ):
    #     refs = ['chr1', 'chr2']
    #     tmpfile = "tmp_%i" % id(self)
    #     s = pysam.AlignmentFile(tmpfile, 'wb',
    #                       referencenames=refs,
    #                       referencelengths=[100]*len(refs))
    #     s.close()

    #     self.assertTrue( checkBinaryEqual( 'issue144.bam', tmpfile ),
    #                      'bam files differ')
    #     os.unlink( tmpfile )



class TestHeaderWriteRead(unittest.TestCase):
    header = {'SQ': [{'LN': 1575, 'SN': 'chr1'},
                     {'LN': 1584, 'SN': 'chr2'}],
              'RG': [{'LB': 'SC_1', 'ID': 'L1', 'SM': 'NA12891',
                      'PU': 'SC_1_10', "CN": "name:with:colon"},
                     {'LB': 'SC_2', 'ID': 'L2', 'SM': 'NA12891',
                      'PU': 'SC_2_12', "CN": "name:with:colon"}],
              'PG': [{'ID': 'P1', 'VN': '1.0', 'CL': 'tool'},
                     {'ID': 'P2', 'VN': '1.1', 'CL': 'tool with in option -R a\tb',
                      'PP': 'P1'}],
              'HD': {'VN': '1.0'},
              'CO': ['this is a comment', 'this is another comment'],
              }

    def compare_headers(self, a, header_b):
        '''compare two headers a and b.

        Ignore M5 and UR field as they are set application specific.
        '''
        b = header_b.to_dict()
        for ak, av in a.items():
            self.assertTrue(ak in b, "key '%s' not in '%s' " % (ak, b))
            self.assertEqual(
                len(av), len(b[ak]),
                "unequal number of entries for key {}: {} vs {}"
                .format(ak, av, b[ak]))

            for row_a, row_b in zip(av, b[ak]):
                if isinstance(row_b, dict):
                    for x in ["M5", "UR"]:
                        try:
                            del row_b[x]
                        except KeyError:
                            pass
                self.assertEqual(row_a, row_b)

    def check_read_write(self, flag_write, header):

        fn = get_temp_filename()
        print(fn)
        with pysam.AlignmentFile(
                fn,
                flag_write,
                header=header,
                reference_filename=os.path.join(BAM_DATADIR, "ex1.fa")) as outf:
            a = pysam.AlignedSegment()
            a.query_name = "abc"
            outf.write(a)

        with pysam.AlignmentFile(fn) as inf:
            read_header = inf.header

        # os.unlink(fn)
        self.compare_headers(header, read_header)
        expected_lengths = dict([(x["SN"], x["LN"]) for x in header["SQ"]])
        self.assertEqual(expected_lengths,
                         dict(zip(read_header.references,
                                  read_header.lengths)))

    def test_SAM(self):
        self.check_read_write("wh", self.header)

    def test_BAM(self):
        self.check_read_write("wb", self.header)
        
    def test_CRAM(self):
        header = copy.copy(self.header)
        if "PG" in header:
            # for CRAM, \t needs to be quoted:
            header['PG'][1]['CL'] = re.sub(r"\t", r"\\\\t", header['PG'][1]['CL'])
        self.check_read_write("wc", header)


class TestHeaderLargeContigs(TestHeaderWriteRead):
    """see issue 741"""

    header = {'SQ': [{'LN': 2147483647, 'SN': 'chr1'},
                     {'LN': 1584, 'SN': 'chr2'}],
              'HD': {'VN': '1.0'}}



"""Benchmarking module for AlignmentFile functionality"""
import os

from TestUtils import BAM_DATADIR, force_str, flatten_nested_list
from PileupTestUtils import *


def test_build_pileup_from_bam_with_samtoolsshell(benchmark):
    result = benchmark(build_pileup_with_samtoolsshell,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 2998


def test_build_pileup_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_pileup_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 2998


def test_build_pileup_from_bam_with_pysam(benchmark):
    result = benchmark(build_pileup_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 2998


def test_build_depth_from_bam_with_samtoolsshell(benchmark):
    result = benchmark(build_depth_with_samtoolsshell,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 107241


def test_build_depth_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_depth_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert sum(result) == 107241


def test_build_depth_from_bam_with_pysam(benchmark):
    result = benchmark(build_depth_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    # different value, as samtools filters with a minimum
    # base quality of 13
    assert sum(result) == 110015


def test_build_depth_with_filter_from_bam_with_pysam(benchmark):
    result = benchmark(build_depth_with_filter_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert sum(result) == 107241


def test_build_query_bases_from_bam_with_samtoolsshell(benchmark):
    result = benchmark(build_query_bases_with_samtoolsshell,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert result == 116308


def test_build_query_bases_from_bam_with_samtoolspysam(benchmark):
    result = benchmark(build_query_bases_with_samtoolspysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join(flatten_nested_list(result))) == 116308
    

def test_build_query_bases_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_query_bases_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join(flatten_nested_list(result))) == 116308


def test_build_query_bases_from_bam_with_pysam_pileups(benchmark):
    # note that there is no overlap detection here
    result = benchmark(build_query_bases_with_pysam_pileups,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join(flatten_nested_list(result))) == 107241


def test_build_query_bases_from_bam_with_pysam(benchmark):
    result = benchmark(build_query_bases_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join(flatten_nested_list(result))) == 116308


# note that pileups with/without reference sequence will differ due to
# realignment.
def test_build_query_bases_with_reference_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_query_bases_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"),
                       "-f", os.path.join(BAM_DATADIR, "ex1.fa"))
    assert len("".join(flatten_nested_list(result))) == 115924
    

def test_build_query_bases_with_reference_from_bam_with_pysam(benchmark):
    with pysam.FastaFile(os.path.join(BAM_DATADIR, "ex1.fa")) as fasta:
        result = benchmark(build_query_bases_with_pysam,
                           os.path.join(BAM_DATADIR, "ex2.bam"),
                           fastafile=fasta)
    assert len("".join(flatten_nested_list(result))) == 115924
    

def test_build_query_bases_with_reference_from_bam_with_samtoolspysam(benchmark):
    result = benchmark(build_query_bases_with_samtoolspysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"),
                       "-f", os.path.join(BAM_DATADIR, "ex1.fa"))
    assert len("".join(flatten_nested_list(result))) == 115924


def test_build_query_qualities_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_query_qualities_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join(result)) == 107241


def test_build_query_qualities_from_bam_with_pysam(benchmark):
    result = benchmark(build_query_qualities_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert sum([len(x) for x in result]) == 107241


def test_build_query_names_from_bam_with_pysam(benchmark):
    result = benchmark(build_query_names_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join([x for column in result for x in column])) == 2307343


def test_build_mapping_qualities_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_mapping_qualities_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert len("".join(result)) == 107241


def test_build_mapping_qualities_from_bam_with_pysam(benchmark):
    result = benchmark(build_mapping_qualities_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert sum([len(x) for x in result]) == 107241


def test_build_query_positions_from_bam_with_samtoolspipe(benchmark):
    result = benchmark(build_query_positions_with_samtoolspipe,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    # positions output by samtools are 1-based
    assert sum([sum(x) - len(x) for x in result]) == 1841699


def test_build_query_positions_from_bam_with_pysam(benchmark):
    result = benchmark(build_query_positions_with_pysam,
                       os.path.join(BAM_DATADIR, "ex2.bam"))
    assert sum([sum(x) for x in result]) == 1841699



"""Functions that read and write block gzipped files.

The user of the file doesn't have to worry about the compression
and random access is allowed if an index file is present."""

# based on Python 3.5's gzip module

import io

from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t
from libc.stdlib cimport malloc, calloc, realloc, free

from cpython.object cimport PyObject
from cpython.bytes  cimport PyBytes_FromStringAndSize, _PyBytes_Resize

from pysam.libcutils   cimport force_bytes, encode_filename
from pysam.libchtslib  cimport bgzf_open, bgzf_index_build_init, bgzf_write, bgzf_read, \
                               bgzf_flush, bgzf_index_dump, bgzf_close, bgzf_seek, \
                               bgzf_tell, bgzf_getline, kstring_t, SEEK_SET, BGZF

__all__ = ["BGZFile"]


BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE


cdef class BGZFile(object):
    """The BGZFile class simulates most of the methods of a file object with
    the exception of the truncate() method.

    This class only supports opening files in binary mode. If you need to open a
    compressed file in text mode, use the gzip.open() function.
    """
    cdef BGZF* bgzf
    cdef readonly object name, index

    def __init__(self, filename, mode=None, index=None):
        """Constructor for the BGZFile class.

        The mode argument can be any of 'r', 'rb', 'a', 'ab', 'w', 'wb', 'x', or
        'xb' depending on whether the file will be read or written.  The default
        is the mode of fileobj if discernible; otherwise, the default is 'rb'.
        A mode of 'r' is equivalent to one of 'rb', and similarly for 'w' and
        'wb', 'a' and 'ab', and 'x' and 'xb'.
        """
        if mode and ('t' in mode or 'U' in mode):
            raise ValueError("Invalid mode: {!r}".format(mode))
        if not mode:
            mode = 'rb'
        elif mode and 'b' not in mode:
            mode += 'b'

        mode = force_bytes(mode)

        self.name = encode_filename(filename)
        self.index = encode_filename(index) if index is not None else None

        self.bgzf = bgzf_open(self.name, mode)

        if self.bgzf.is_write and index is not None and bgzf_index_build_init(self.bgzf) < 0:
            raise IOError('Error building bgzf index')

    def __dealloc__(self):
        self.close()

    def write(self, data):
        if not self.bgzf:
            raise ValueError("write() on closed BGZFile object")

        if not self.bgzf.is_write:
            import errno
            raise IOError(errno.EBADF, "write() on read-only BGZFile object")

        if isinstance(data, bytes):
            length = len(data)
        else:
            # accept any data that supports the buffer protocol
            data = memoryview(data)
            length = data.nbytes

        if length > 0 and bgzf_write(self.bgzf, <char *>data, length) < 0:
            raise IOError('BGZFile write failed')

        return length

    def read(self, size=-1):
        cdef ssize_t read_size

        if not self.bgzf:
            raise ValueError("read() on closed BGZFile object")

        if self.bgzf.is_write:
            import errno
            raise IOError(errno.EBADF, "read() on write-only BGZFile object")

        if size < 0:
            chunks = []
            while 1:
                chunk = PyBytes_FromStringAndSize(NULL, BUFFER_SIZE)
                cdata = <bytes>chunk
                read_size = bgzf_read(self.bgzf, <char *>chunk, BUFFER_SIZE)
                if read_size < 0:
                    raise IOError('Error reading from BGZFile')
                elif not read_size:
                    break
                elif read_size < BUFFER_SIZE:
                    chunk = chunk[:read_size]
                chunks.append(chunk)
            return b''.join(chunks)

        elif size > 0:
            chunk = PyBytes_FromStringAndSize(NULL, size)
            read_size = bgzf_read(self.bgzf, <char *>chunk, size)
            if read_size < 0:
                raise IOError('Error reading from BGZFile')
            elif read_size < size:
                chunk = chunk[:size]
            return chunk
        else:
            return b''

    @property
    def closed(self):
        return self.bgzf == NULL

    def close(self):
        if not self.bgzf:
            return

        if self.bgzf.is_write and bgzf_flush(self.bgzf) < 0:
            raise IOError('Error flushing BGZFile object')

        if self.index and bgzf_index_dump(self.bgzf, self.index, NULL) < 0:
            raise IOError('Cannot write index')

        cdef ret = bgzf_close(self.bgzf)
        self.bgzf = NULL

        if ret < 0:
            raise IOError('Error closing BGZFile object')

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def flush(self):
        if not self.bgzf:
            return

        if self.bgzf.is_write and bgzf_flush(self.bgzf) < 0:
            raise IOError('Error flushing BGZFile object')

    def fileno(self):
        """Invoke the underlying file object's fileno() method.

        This will raise AttributeError if the underlying file object
        doesn't support fileno().
        """
        raise AttributeError('fileno')

    def rewind(self):
        '''Return the uncompressed stream file position indicator to the
        beginning of the file'''
        if not self.bgzf:
            raise ValueError("rewind() on closed BGZFile object")
        if not self.bgzf.is_write:
            raise IOError("Can't rewind in write mode")
        if bgzf_seek(self.bgzf, 0, SEEK_SET) < 0:
            raise IOError('Error seeking BGZFFile object')

    def readable(self):
        if not self.bgzf:
            raise ValueError("readable() on closed BGZFile object")
        return self.bgzf != NULL and not self.bgzf.is_write

    def writable(self):
        return self.bgzf != NULL and self.bgzf.is_write

    def seekable(self):
        return True

    def tell(self):
        if not self.bgzf:
            raise ValueError("seek() on closed BGZFile object")
        cdef int64_t off = bgzf_tell(self.bgzf)
        if off < 0:
            raise IOError('Error in tell on BGZFFile object')

        return off

    def seek(self, offset, whence=io.SEEK_SET):
        if not self.bgzf:
            raise ValueError("seek() on closed BGZFile object")
        if whence is not io.SEEK_SET:
            raise ValueError('Seek from end not supported')

        cdef int64_t off = bgzf_seek(self.bgzf, offset, SEEK_SET)
        if off < 0:
            raise IOError('Error seeking BGZFFile object')

        return off

    def readline(self, size=-1):
        if not self.bgzf:
            raise ValueError("readline() on closed BGZFile object")

        cdef kstring_t line
        cdef char c

        line.l = line.m = 0
        line.s = NULL

        cdef int ret = bgzf_getline(self.bgzf, '\n', &line)
        if ret == -1:
            s = b''
        elif ret == -2:
            if line.m:
                free(line.s)
            raise IOError('Error reading line in BGZFFile object')
        else:
            s = line.s[:line.l]

        if line.m:
            free(line.s)

        return s

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration()
        return line



# cython: embedsignature=True
# cython: profile=True
# adds doc-strings for sphinx
import tempfile
import os
import sys
import types
import itertools
import struct
import ctypes
import collections
import re
import platform
import warnings
from cpython cimport PyErr_SetString, \
    PyBytes_Check, \
    PyUnicode_Check, \
    PyBytes_FromStringAndSize

from cpython.version cimport PY_MAJOR_VERSION

from pysam.libcalignmentfile cimport AlignmentFile, AlignedSegment


cdef class Samfile(AlignmentFile):
    '''Deprecated alternative for :class:`~pysam.AlignmentFile`

    Added for backwards compatibility with pysam <= 0.8.0
    '''
    pass


cdef class AlignedRead(AlignedSegment):
    '''Deprecated alternative for :class:`~pysam.AlignedSegment`

    Added for backwards compatibility with pysam <= 0.8.0
    '''
    pass


__all__ = ['Samfile', 'AlignedRead']


